@@ -6,36 +6,37 @@
 /*                                                                     */
 /*  Copyright 1996 Institut National de Recherche en Informatique et   */
 /*  en Automatique.  All rights reserved.  This file is distributed    */
-/*  under the terms of the GNU Library General Public License.         */
+/*  under the terms of the GNU Library General Public License, with    */
+/*  the special exception on linking described in file ../LICENSE.     */
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: config.h,v 1.21 2000/04/05 18:30:11 doligez Exp $ */
+/* $Id: config.h,v 1.30 2002/08/01 14:14:10 doligez Exp $ */
 
 #ifndef _config_
 #define _config_
 
-
-#if !macintosh
-#undef ARCH_SIXTYFOUR
+#define ARCH_SIXTYFOUR
 #define SIZEOF_INT 4
-#define SIZEOF_LONG 4
+#define SIZEOF_LONG 8
 #define SIZEOF_SHORT 2
-#define ARCH_INT64_TYPE long long
-#define ARCH_UINT64_TYPE unsigned long long
-#define ARCH_INT64_PRINTF_FORMAT "ll"
+#define ARCH_INT64_TYPE long
+#define ARCH_UINT64_TYPE unsigned long
+#define ARCH_INT64_PRINTF_FORMAT "l"
 #define ARCH_BIG_ENDIAN
-#define ARCH_ALIGN_DOUBLE
-#define ARCH_ALIGN_INT64
+#undef ARCH_ALIGN_DOUBLE
+#undef ARCH_ALIGN_INT64
+#undef NONSTANDARD_DIV_MOD
 #define OCAML_OS_TYPE "Unix"
-#define HAS_MEMMOVE
-#define HAS_BCOPY
+#define OCAML_STDLIB_DIR "/usr/local/lib/ocaml"
 #define POSIX_SIGNALS
 #define HAS_STRERROR
 #define HAS_TIMES
 #define HAS_TERMCAP
 #define HAS_SOCKETS
+#define HAS_INET_ATON
 #define HAS_UNISTD
+#define HAS_OFF_T
 #define HAS_DIRENT
 #define HAS_REWINDDIR
 #define HAS_LOCKF
@@ -48,8 +49,8 @@
 #define HAS_DUP2
 #define HAS_FCHMOD
 #define HAS_TRUNCATE
-#define HAS_SELECT
 #define HAS_SYS_SELECT_H
+#define HAS_SELECT
 #define HAS_SYMLINK
 #define HAS_WAITPID
 #define HAS_WAIT4
@@ -64,10 +65,6 @@
 #define HAS_PUTENV
 #define HAS_LOCALE
 #define HAS_MMAP
-#else
-#include <m.h>
-#include <s.h>
-#endif
 
 /* Types for signed chars, 16-bit integers, 32-bit integers, 64-bit integers */
 
@@ -87,37 +84,34 @@ typedef short int32;
 typedef unsigned short uint32;
 #endif
 
-#if defined(ARCH_INT64_TYPE) && defined(ARCH_UINT64_TYPE)
+#if defined(ARCH_INT64_TYPE)
 typedef ARCH_INT64_TYPE int64;
 typedef ARCH_UINT64_TYPE uint64;
 #else
-/* Int64.t will not be supported, and operations over it are not defined,
-   but we must define the types int64 and uint64 as 64-bit placeholders. */
-typedef struct { uint32 a, b; } uint64;
-typedef uint64 int64;
+#  if ARCH_BIG_ENDIAN
+typedef struct { uint32 h, l; } uint64, int64;
+#  else
+typedef struct { uint32 l, h; } uint64, int64;
+#  endif
 #endif
 
+/* Endianness of floats */
 
-/* Library dependencies */
+/* ARCH_FLOAT_ENDIANNESS encodes the byte order of doubles as follows:
+   the value [0xabcdefgh] means that the least significant byte of the
+   float is at byte offset [a], the next lsb at [b], ..., and the
+   most significant byte at [h]. */
 
-#ifdef HAS_MEMMOVE
-#undef bcopy
-#define bcopy(src,dst,len) memmove((dst), (src), (len))
+#if defined(__arm__)
+#define ARCH_FLOAT_ENDIANNESS 0x45670123
+#elif defined(ARCH_BIG_ENDIAN)
+#define ARCH_FLOAT_ENDIANNESS 0x76543210
 #else
-#ifdef HAS_BCOPY
-/* Nothing to do */
-#else
-#undef bcopy
-#define bcopy(src,dst,len) memmov((dst), (src), (len))
-#define USING_MEMMOV
-#endif
+#define ARCH_FLOAT_ENDIANNESS 0x01234567
 #endif
 
 /* We use threaded code interpretation if the compiler provides labels
-   as first-class values (GCC 2.x).
-   Macintosh 68k also uses threaded code, with the assembly-language
-   bytecode interpreter (THREADED_CODE defined in config/sm-Mac.h).
-*/
+   as first-class values (GCC 2.x). */
 
 #if defined(__GNUC__) && __GNUC__ >= 2 && !defined(DEBUG) && !defined (SHRINKED_GNUC)
 #define THREADED_CODE
@@ -181,11 +175,15 @@ typedef uint64 int64;
 
 /* Default speed setting for the major GC.  The heap will grow until
    the dead objects and the free list represent this percentage of the
-   heap size.  The rest of the heap is live objects. */
-#define Percent_free_def 42
+   total size of live objects. */
+#define Percent_free_def 80
 
-/* Default setting for the compacter: off */
-#define Max_percent_free_def 1000000
+/* Default setting for the compacter: 500%
+   (i.e. trigger the compacter when 5/6 of the heap is free or garbage)
+   This can be set quite high because the overhead is over-estimated
+   when fragmentation occurs.
+ */
+#define Max_percent_free_def 500
 
 
 #endif /* _config_ */
