@@ -6,11 +6,12 @@
 /*                                                                     */
 /*  Copyright 1996 Institut National de Recherche en Informatique et   */
 /*  en Automatique.  All rights reserved.  This file is distributed    */
-/*  under the terms of the GNU Library General Public License.         */
+/*  under the terms of the GNU Library General Public License, with    */
+/*  the special exception on linking described in file ../LICENSE.     */
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: memory.h,v 1.30 2000/04/03 08:34:21 doligez Exp $ */
+/* $Id: memory.h,v 1.40 2002/06/03 14:21:50 xleroy Exp $ */
 
 /* Allocation macros and functions */
 
@@ -22,40 +23,21 @@
 #include "misc.h"
 #include "mlvalues.h"
 
-value alloc_shr (mlsize_t, tag_t);
+CAMLextern value alloc_shr (mlsize_t, tag_t);
 void adjust_gc_speed (mlsize_t, mlsize_t);
-void modify (value *, value);
-void initialize (value *, value);
-value check_urgent_gc (value);
-void * stat_alloc (asize_t);              /* Size in bytes. */
-void stat_free (void *);
-void * stat_resize (void *, asize_t);     /* Size in bytes. */
-header_t *alloc_for_heap (asize_t request);   /* Size in bytes. */
-void free_for_heap (header_t *mem);
-int add_to_heap (header_t *mem);
+CAMLextern void modify (value *, value);
+CAMLextern void initialize (value *, value);
+CAMLextern value check_urgent_gc (value);
+CAMLextern void * stat_alloc (asize_t);              /* Size in bytes. */
+CAMLextern void stat_free (void *);
+CAMLextern void * stat_resize (void *, asize_t);     /* Size in bytes. */
+char *alloc_for_heap (asize_t request);   /* Size in bytes. */
+void free_for_heap (char *mem);
+int add_to_heap (char *mem);
 color_t allocation_color (void *hp);
 
 /* void shrink_heap (char *);        Only used in compact.c */
 
-#ifdef NATIVE_CODE
-#define Garbage_collection_function garbage_collection
-#else
-#define Garbage_collection_function minor_collection
-#endif
-
-#ifdef DEBUG
-#define DEBUG_clear(result, wosize) { \
-  unsigned long __DEBUG_i; \
-  for (__DEBUG_i = 0; __DEBUG_i < wosize; ++ __DEBUG_i){ \
-    Field (result, __DEBUG_i) = Debug_uninit_minor; \
-  } \
-}
-#else
-#define DEBUG_clear(result, wosize)
-#endif
-
-
-
 
 struct caml__roots_block {
   struct caml__roots_block *next;
@@ -64,7 +46,7 @@ struct caml__roots_block {
   value *tables [5];
 };
 
-extern struct caml__roots_block *local_roots;  /* defined in roots.c */
+CAMLextern struct caml__roots_block *local_roots;  /* defined in roots.c */
 
 /* The following macros are used to declare C local variables and
    function parameters of type [value].
@@ -85,7 +67,7 @@ extern struct caml__roots_block *local_roots;  /* defined in roots.c */
    Use [CAMLlocalN] to declare an array of [value]s.
 
    Your function may raise and exception or return a [value] with the
-   [CAMLreturn1] macro.  Its argument is simply the [value] returned by
+   [CAMLreturn] macro.  Its argument is simply the [value] returned by
    your function.  Do NOT directly return a [value] with the [return]
    keyword.  If your function returns void, use [CAMLreturn0].
 
@@ -228,7 +210,11 @@ extern struct caml__roots_block *local_roots;  /* defined in roots.c */
 }while(0)
 
 /* convenience macro */
-#define Store_field(block, offset, val) modify (&Field (block, offset), val)
+#define Store_field(block, offset, val) do{ \
+  mlsize_t caml__temp_offset = (offset); \
+  value caml__temp_val = (val); \
+  modify (&Field ((block), caml__temp_offset), caml__temp_val); \
+}while(0)
 
 /*
    NOTE: [Begin_roots] and [End_roots] are superseded by [CAMLparam]*,
@@ -317,12 +303,12 @@ extern struct caml__roots_block *local_roots;  /* defined in roots.c */
    for the duration of the program, or until [remove_global_root] is
    called. */
 
-void register_global_root (value *);
+CAMLextern void register_global_root (value *);
 
 /* [remove_global_root] removes a memory root registered on a global C
    variable with [register_global_root]. */
 
-void remove_global_root (value *);
+CAMLextern void remove_global_root (value *);
 
 
 #endif /* _memory_ */
