@@ -6,26 +6,37 @@
 (*                                                                     *)
 (*  Copyright 1996 Institut National de Recherche en Informatique et   *)
 (*  en Automatique.  All rights reserved.  This file is distributed    *)
-(*  under the terms of the GNU Library General Public License.         *)
+(*  under the terms of the GNU Library General Public License, with    *)
+(*  the special exception on linking described in file ../LICENSE.     *)
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: format.ml,v 1.28 1999/11/17 18:58:23 xleroy Exp $ *)
+(* $Id: format.ml,v 1.46 2002/05/15 20:29:26 weis Exp $ *)
+
+(**************************************************************
+
+  Data structures definitions.
+
+ **************************************************************)
 
 (* Tokens are one of the following : *)
 
-type pp_token = 
+type pp_token =
 | Pp_text of string            (* normal text *)
 | Pp_break of int * int        (* complete break *)
-| Pp_tbreak of int * int       (* go to next tab *)
+| Pp_tbreak of int * int       (* go to next tabulation *)
 | Pp_stab                      (* set a tabulation *)
 | Pp_begin of int * block_type (* beginning of a block *)
 | Pp_end                       (* end of a block *)
-| Pp_tbegin of tblock          (* Beginning of a tabulation block *)
+| Pp_tbegin of tblock          (* beginning of a tabulation block *)
 | Pp_tend                      (* end of a tabulation block *)
 | Pp_newline                   (* to force a newline inside a block *)
 | Pp_if_newline                (* to do something only if this very
                                   line has been broken *)
+| Pp_open_tag of string        (* opening a tag name *)
+| Pp_close_tag                 (* closing the most recently opened tag *)
+
+and tag = string
 
 and block_type =
 | Pp_hbox   (* Horizontal block no line breaking *)
@@ -42,19 +53,19 @@ and block_type =
 and tblock = Pp_tbox of int list ref  (* Tabulation box *)
 ;;
 
-(* The Queue: contains all formatting elements.
+(* The Queue:
+   contains all formatting elements.
    elements are tuples (size, token, length), where
    size is set when the size of the block is known
    len is the declared length of the token *)
-type pp_queue_elem =
-{mutable elem_size : int; token : pp_token; length : int};;
+type pp_queue_elem = {mutable elem_size : int; token : pp_token; length : int};;
 
-(* Scan stack
+(* Scan stack:
    each element is (left_total, queue element) where left_total
    is the value of pp_left_total when the element has been enqueued *)
 type pp_scan_elem = Scan_elem of int * pp_queue_elem;;
 
-(* Formatting Stack:
+(* Formatting stack:
    used to break the lines while printing tokens.
    The formatting stack contains the description of
    the currently active blocks. *)
@@ -64,14 +75,27 @@ type pp_format_elem = Format_elem of block_type * int;;
 type 'a queue_elem = | Nil | Cons of 'a queue_cell
 and 'a queue_cell = {mutable head : 'a; mutable tail : 'a queue_elem};;
 
-type 'a queue =
-{mutable insert : 'a queue_elem;
- mutable body : 'a queue_elem};;
+type 'a queue = {
+ mutable insert : 'a queue_elem;
+ mutable body : 'a queue_elem
+};;
 
-type formatter =
-{mutable pp_scan_stack : pp_scan_elem list;
+(* The formatter specific tag handling functions. *)
+type formatter_tag_functions = {
+ mark_open_tag : tag -> string;
+ mark_close_tag : tag -> string;
+ print_open_tag : tag -> unit;
+ print_close_tag : tag -> unit;
+
+};;
+
+(* A formatter with all its machinery. *)
+type formatter = {
+ mutable pp_scan_stack : pp_scan_elem list;
  mutable pp_format_stack : pp_format_elem list;
  mutable pp_tbox_stack : tblock list;
+ mutable pp_tag_stack : tag list;
+ mutable pp_mark_stack : tag list;
  (* Global variables: default initialization is
     set_margin 78
     set_min_space_left 0 *)
@@ -103,14 +127,30 @@ type formatter =
  (* Flushing function *)
  mutable pp_flush_function : unit -> unit;
  (* Output of new lines *)
- mutable pp_output_newline : formatter -> unit -> unit;
+ mutable pp_output_newline : unit -> unit;
  (* Output of indentation spaces *)
- mutable pp_output_spaces : formatter -> int -> unit;
+ mutable pp_output_spaces : int -> unit;
+ (* Are tags printed ? *)
+ mutable pp_print_tags : bool;
+ (* Are tags marked ? *)
+ mutable pp_mark_tags : bool;
+ (* Find opening and closing markers of tags. *)
+ mutable pp_mark_open_tag : tag -> string;
+ mutable pp_mark_close_tag : tag -> string;
+ mutable pp_print_open_tag : tag -> unit;
+ mutable pp_print_close_tag : tag -> unit;
  (* The pretty-printer queue *)
  mutable pp_queue : pp_queue_elem queue
 };;
 
-(* Qeues *)
+(**************************************************************
+
+  Auxilliaries and basic functions.
+
+ **************************************************************)
+
+
+(* Qeues auxilliaries. *)
 let make_queue () = {insert = Nil; body = Nil};;
 
 let clear_queue q = q.insert <- Nil; q.body <- Nil;;
@@ -151,9 +191,9 @@ let pp_infinity = 999999999;;
 
 (* Output functions for the formatter *)
 let pp_output_string state s = state.pp_output_function s 0 (String.length s)
-and pp_output_newline state = state.pp_output_newline state ();;
+and pp_output_newline state = state.pp_output_newline ();;
 
-let pp_display_blanks state n = state.pp_output_spaces state n;;
+let pp_display_blanks state n = state.pp_output_spaces n;;
 
 (* To format a break, indenting a new line *)
 let break_new_line state offset width =
@@ -193,6 +233,12 @@ let pp_skip_token state =
        state.pp_left_total <- state.pp_left_total - len;
        state.pp_space_left <- state.pp_space_left + size;;
 
+(**************************************************************
+
+  The main pretting printing functions.
+
+ **************************************************************)
+
 (* To format a token *)
 let format_pp_token state size = function
 
@@ -232,7 +278,7 @@ let format_pp_token state size = function
 
   | Pp_stab ->
      begin match state.pp_tbox_stack with
-     | Pp_tbox tabs :: _ -> 
+     | Pp_tbox tabs :: _ ->
         let rec add_tab n = function
           | [] -> [n]
           | x :: l as ls -> if n < x then n :: ls else x :: add_tab n l in
@@ -243,7 +289,7 @@ let format_pp_token state size = function
   | Pp_tbreak (n, off) ->
       let insertion_point = state.pp_margin - state.pp_space_left in
       begin match state.pp_tbox_stack with
-      | Pp_tbox tabs :: _ -> 
+      | Pp_tbox tabs :: _ ->
          let rec find n = function
            | x :: l -> if x >= n then x else find n l
            | [] -> raise Not_found in
@@ -273,7 +319,7 @@ let format_pp_token state size = function
      | Format_elem (ty, width) :: _ ->
         begin match ty with
         | Pp_hovbox ->
-           if size > state.pp_space_left 
+           if size > state.pp_space_left
            then break_new_line state off width
            else break_same_line state n
         | Pp_box ->
@@ -283,14 +329,29 @@ let format_pp_token state size = function
             then break_new_line state off width else
            (* break the line here leads to new indentation ? *)
            if state.pp_current_indent > state.pp_margin - width + off
-           then break_new_line state off width else break_same_line state n
+           then break_new_line state off width
+           else break_same_line state n
         | Pp_hvbox -> break_new_line state off width
         | Pp_fits -> break_same_line state n
-        | Pp_vbox  -> break_new_line state off width
-        | Pp_hbox  -> break_same_line state n
+        | Pp_vbox -> break_new_line state off width
+        | Pp_hbox -> break_same_line state n
         end
      | _ -> () (* No opened block *)
-     end;;
+     end
+
+   | Pp_open_tag tag_name ->
+      let marker = state.pp_mark_open_tag tag_name in
+      pp_output_string state marker;
+      state.pp_mark_stack <- tag_name :: state.pp_mark_stack
+
+   | Pp_close_tag ->
+      begin match state.pp_mark_stack with
+      | tag_name :: tags ->
+          let marker = state.pp_mark_close_tag tag_name in
+          pp_output_string state marker;
+          state.pp_mark_stack <- tags
+      | _ -> () (* No more tag to close *)
+      end;;
 
 (* Print if token size is known or printing is delayed
    Size is known when not negative
@@ -365,11 +426,9 @@ let scan_push state b tok =
     state.pp_scan_stack <-
      Scan_elem (state.pp_right_total, tok) :: state.pp_scan_stack;;
 
-(*
-  To open a new block :
-  the user may set the depth bound pp_max_boxes
-  any text nested deeper is printed as the ellipsis string
-*)
+(* To open a new block :
+   the user may set the depth bound pp_max_boxes
+   any text nested deeper is printed as the ellipsis string *)
 let pp_open_box_gen state indent br_ty =
     state.pp_curr_depth <- state.pp_curr_depth + 1;
     if state.pp_curr_depth < state.pp_max_boxes then
@@ -398,6 +457,50 @@ let pp_close_box state () =
       state.pp_curr_depth <- state.pp_curr_depth - 1;
      end;;
 
+(* Open a tag, pushing it on the tag stack. *)
+let pp_open_tag state tag_name =
+    if state.pp_print_tags then begin
+      state.pp_tag_stack <- tag_name :: state.pp_tag_stack;
+      state.pp_print_open_tag tag_name end;
+    if state.pp_mark_tags then
+      pp_enqueue state
+        {elem_size = 0; token = Pp_open_tag tag_name; length = 0};;
+
+(* Close a tag, popping it from the tag stack. *)
+let pp_close_tag state () =
+    if state.pp_mark_tags then
+      pp_enqueue state {elem_size = 0; token = Pp_close_tag; length = 0};
+    if state.pp_print_tags then
+      begin match state.pp_tag_stack with
+      | tag_name :: tags ->
+          state.pp_print_close_tag tag_name;
+          state.pp_tag_stack <- tags
+      | _ -> () (* No more tag to close *)
+      end;;
+
+let pp_set_print_tags state b = state.pp_print_tags <- b;;
+let pp_set_mark_tags state b = state.pp_mark_tags <- b;;
+let pp_get_print_tags state () = state.pp_print_tags;;
+let pp_get_mark_tags state () = state.pp_mark_tags;;
+let pp_set_tags state b = pp_set_print_tags state b; pp_set_mark_tags state b;;
+
+let pp_get_formatter_tag_functions state () = {
+   mark_open_tag = state.pp_mark_open_tag;
+   mark_close_tag = state.pp_mark_close_tag;
+   print_open_tag = state.pp_print_open_tag;
+   print_close_tag = state.pp_print_close_tag;
+};;
+
+let pp_set_formatter_tag_functions state
+    {mark_open_tag = motag;
+     mark_close_tag = mctag;
+     print_open_tag = potag;
+     print_close_tag = pctag;} =
+  state.pp_mark_open_tag <- motag;
+  state.pp_mark_close_tag <- mctag;
+  state.pp_print_open_tag <- potag;
+  state.pp_print_close_tag <- pctag;;
+
 (* Initialize pretty-printer. *)
 let pp_rinit state =
     pp_clear_queue state;
@@ -407,6 +510,8 @@ let pp_rinit state =
     state.pp_space_left <- state.pp_margin;
     state.pp_format_stack <- [];
     state.pp_tbox_stack <- [];
+    state.pp_tag_stack <- [];
+    state.pp_mark_stack <- [];
     pp_open_sys_box state;;
 
 (* Flushing pretty-printer queue. *)
@@ -474,7 +579,7 @@ let pp_print_if_newline state () =
    block else (the value of) width blanks are printed.
    To do (?) : add a maximum width and offset value *)
 let pp_print_break state width offset =
-  if state.pp_curr_depth < state.pp_max_boxes then 
+  if state.pp_curr_depth < state.pp_max_boxes then
     scan_push state true
      {elem_size = (- state.pp_right_total); token = Pp_break (width, offset);
       length = width};;
@@ -501,7 +606,7 @@ let pp_close_tbox state () =
 let pp_print_tbreak state width offset =
   if state.pp_curr_depth < state.pp_max_boxes then
     scan_push state true
-     {elem_size = (- state.pp_right_total); token = Pp_tbreak (width, offset); 
+     {elem_size = (- state.pp_right_total); token = Pp_tbreak (width, offset);
       length = width};;
 
 let pp_print_tab state () = pp_print_tbreak state 0 0;;
@@ -512,7 +617,7 @@ let pp_set_tab state () =
 
 (**************************************************************
 
-  Procedures to control the pretty-printer
+  Procedures to control the pretty-printers
 
  **************************************************************)
 
@@ -528,7 +633,7 @@ let pp_over_max_boxes state () = state.pp_curr_depth = state.pp_max_boxes;;
 let pp_set_ellipsis_text state s = state.pp_ellipsis <- s
 and pp_get_ellipsis_text state () = state.pp_ellipsis;;
 
-(* To set the margin of pretty-formater *)
+(* To set the margin of pretty-printer *)
 let pp_set_min_space_left state n =
   if n >= 1 && n < pp_infinity then
    begin
@@ -536,7 +641,7 @@ let pp_set_min_space_left state n =
     state.pp_max_indent <- state.pp_margin - state.pp_min_space_left;
     pp_rinit state end;;
 
-(* Initially we have :
+(* Initially, we have :
   pp_max_indent = pp_margin - pp_min_space_left, and
   pp_space_left = pp_margin *)
 let pp_set_max_indent state n =
@@ -563,22 +668,35 @@ let pp_get_margin state () = state.pp_margin;;
 
 let pp_set_formatter_output_functions state f g =
   state.pp_output_function <- f; state.pp_flush_function <- g;;
-let pp_get_formatter_output_functions state () = 
+let pp_get_formatter_output_functions state () =
   (state.pp_output_function, state.pp_flush_function);;
 
-let pp_set_all_formatter_output_functions state f g h i =
+let pp_set_all_formatter_output_functions state
+    ~out:f ~flush:g ~newline:h ~spaces:i =
   pp_set_formatter_output_functions state f g;
-  state.pp_output_newline <- (function _ -> function () -> h ());
-  state.pp_output_spaces <- (function _ -> function n -> i n);;
-let pp_get_all_formatter_output_functions state () = 
+  state.pp_output_newline <- (function () -> h ());
+  state.pp_output_spaces <- (function n -> i n);;
+let pp_get_all_formatter_output_functions state () =
   (state.pp_output_function, state.pp_flush_function,
-   state.pp_output_newline state, state.pp_output_spaces state);;
+   state.pp_output_newline, state.pp_output_spaces);;
 
 let pp_set_formatter_out_channel state os =
   state.pp_output_function <- output os;
   state.pp_flush_function <- (fun () -> flush os);;
 
-let pp_make_formatter f g h i = 
+(**************************************************************
+
+  Creation of specific formatters
+
+ **************************************************************)
+
+let default_pp_mark_open_tag s = "<" ^ s ^ ">";;
+let default_pp_mark_close_tag s = "</" ^ s ^ ">";;
+
+let default_pp_print_open_tag s = ();;
+let default_pp_print_close_tag = default_pp_print_open_tag;;
+
+let pp_make_formatter f g h i =
  (* The initial state of the formatter contains a dummy box *)
  let pp_q = make_queue () in
  let sys_tok =
@@ -589,6 +707,8 @@ let pp_make_formatter f g h i =
  {pp_scan_stack = sys_scan_stack;
   pp_format_stack = [];
   pp_tbox_stack = [];
+  pp_tag_stack = [];
+  pp_mark_stack = [];
   pp_margin = 78;
   pp_min_space_left = 10;
   pp_max_indent = 78 - 10;
@@ -598,12 +718,18 @@ let pp_make_formatter f g h i =
   pp_left_total = 1;
   pp_right_total = 1;
   pp_curr_depth = 1;
-  pp_max_boxes = 35;
+  pp_max_boxes = max_int;
   pp_ellipsis = ".";
   pp_output_function = f;
   pp_flush_function = g;
   pp_output_newline = h;
   pp_output_spaces = i;
+  pp_print_tags = true;
+  pp_mark_tags = true;
+  pp_mark_open_tag = default_pp_mark_open_tag;
+  pp_mark_close_tag = default_pp_mark_close_tag;
+  pp_print_open_tag = default_pp_print_open_tag;
+  pp_print_close_tag = default_pp_print_close_tag;
   pp_queue = pp_q
  };;
 
@@ -620,13 +746,19 @@ let rec display_blanks state n =
 (* Default function to output new lines *)
 let display_newline state () = state.pp_output_function "\n" 0  1;;
 
-let make_formatter f g = pp_make_formatter f g display_newline display_blanks;;
+let make_formatter f g =
+  let ff = pp_make_formatter f g ignore ignore in
+  ff.pp_output_newline <- display_newline ff;
+  ff.pp_output_spaces <- display_blanks ff;
+  ff;;
 
 let formatter_of_out_channel oc =
- make_formatter (output oc) (fun () -> flush oc);;
+  make_formatter (output oc) (fun () -> flush oc);;
+
+let unit_out () = ();;
 
 let formatter_of_buffer b =
-  make_formatter (Buffer.add_substring b) (fun () -> ());;
+  make_formatter (Buffer.add_substring b) unit_out;;
 
 let stdbuf = Buffer.create 512;;
 
@@ -635,10 +767,16 @@ let std_formatter = formatter_of_out_channel stdout;;
 let err_formatter = formatter_of_out_channel stderr;;
 
 let flush_str_formatter () =
- pp_flush_queue str_formatter false;
- let s = Buffer.contents stdbuf in
- Buffer.reset stdbuf;
- s;;
+  pp_flush_queue str_formatter false;
+  let s = Buffer.contents stdbuf in
+  Buffer.reset stdbuf;
+  s;;
+
+(**************************************************************
+
+  Basic functions on the standard formatter
+
+ **************************************************************)
 
 let open_hbox = pp_open_hbox std_formatter
 and open_vbox = pp_open_vbox std_formatter
@@ -646,6 +784,8 @@ and open_hvbox = pp_open_hvbox std_formatter
 and open_hovbox = pp_open_hovbox std_formatter
 and open_box = pp_open_box std_formatter
 and close_box = pp_close_box std_formatter
+and open_tag = pp_open_tag std_formatter
+and close_tag = pp_close_tag std_formatter
 and print_as = pp_print_as std_formatter
 and print_string = pp_print_string std_formatter
 and print_int = pp_print_int std_formatter
@@ -691,26 +831,106 @@ and get_formatter_output_functions =
 and set_all_formatter_output_functions =
     pp_set_all_formatter_output_functions std_formatter
 and get_all_formatter_output_functions =
-    pp_get_all_formatter_output_functions std_formatter;;
+    pp_get_all_formatter_output_functions std_formatter
 
-external format_int: string -> int -> string = "format_int"
-external format_float: string -> float -> string = "format_float"
+and set_formatter_tag_functions =
+    pp_set_formatter_tag_functions std_formatter
+and get_formatter_tag_functions =
+    pp_get_formatter_tag_functions std_formatter
+and set_print_tags =
+    pp_set_print_tags std_formatter
+and get_print_tags =
+    pp_get_print_tags std_formatter
+and set_mark_tags =
+    pp_set_mark_tags std_formatter
+and get_mark_tags =
+    pp_get_mark_tags std_formatter
+and set_tags =
+    pp_set_tags std_formatter
+;;
 
-let format_invalid_arg s c = invalid_arg (s ^ String.make 1 c);;
 
+(**************************************************************
+
+  Printf implementation.
+
+ **************************************************************)
+
+(* Error messages when processing formats. *)
+
+(* Trailer: giving up at character number ... *)
+let giving_up mess fmt i =
+  "fprintf: " ^ mess ^ " ``" ^ fmt ^ "'', \
+   giving up at character number " ^ string_of_int i ^
+  (if i < String.length fmt
+   then " (" ^ String.make 1 fmt.[i] ^ ")."
+   else String.make 1 '.');;
+
+(* When an invalid format deserve a special error explanation. *)
+let format_invalid_arg mess fmt i = invalid_arg (giving_up mess fmt i);;
+
+(* Standard invalid format. *)
+let invalid_format fmt i = format_invalid_arg "bad format" fmt i;;
+
+(* Cannot find a valid integer into that format. *)
+let invalid_integer fmt i =
+  invalid_arg (giving_up "bad integer specification" fmt i);;
+
+(* Finding an integer out of a sub-string of the format. *)
+let format_int_of_string fmt i s =
+  try int_of_string s with
+  | Failure s -> invalid_integer fmt i;;
+
+(* Getting strings out of buffers. *)
+let get_buffer_out b =
+ let s = Buffer.contents b in
+ Buffer.reset b;
+ s;;
+
+(* [ppf] is supposed to be a pretty-printer that outputs in buffer [b]:
+   to extract contents of [ppf] as a string we flush [ppf] and get the string
+   out of [b]. *)
+let string_out b ppf =
+ pp_flush_queue ppf false;
+ get_buffer_out b;;
+
+(* Applies [printer] to a formatter that outputs on a fresh buffer,
+   then returns the resulting material *)
+let exstring printer arg =
+ let b = Buffer.create 512 in
+ let ppf = formatter_of_buffer b in
+ printer ppf arg;
+ string_out b ppf;;
+
+(* To turn out a character accumulator into the proper string result. *)
+let implode_rev s0 = function
+  | [] -> s0
+  | l -> String.concat "" (List.rev (s0 :: l));;
+
+(* Safe coercion from a format to a string. *)
+external string_of_format : ('a, 'b, 'c) format -> string = "%identity";;
+
+(* [fprintf_out] is the printf-like function generator: given the
+   - [str] flag that tells if we are printing into a string,
+   - the [out] function that has to be called at the end of formatting,
+   it generates a [fprintf] function that takes as arguments a [ppf]
+   formatter and a printing format to print the rest of arguments
+   according to the format.
+   Regular [fprintf]-like functions of this module are obtained via partial
+   applications of [fprintf_out]. *)
 let fprintf_out str out ppf format =
-  let format = (Obj.magic format : string) in
+  let format = string_of_format format in
   let limit = String.length format in
 
   let print_as = ref None in
 
-  let pp_print_as_char ppf c =
+  let pp_print_as_char c =
       match !print_as with
       | None -> pp_print_char ppf c
       | Some size ->
          pp_print_as ppf size (String.make 1 c);
          print_as := None
-  and pp_print_as_string ppf s =
+  and pp_print_as_string s =
       match !print_as with
       | None -> pp_print_string ppf s
       | Some size ->
@@ -722,204 +942,191 @@ let fprintf_out str out ppf format =
       Obj.magic (out ())
     else
       match format.[i] with
+      | '%' ->
+          Printf.scan_format format i cont_s cont_a cont_t
       | '@' ->
-          let j = succ i in
-          if j >= limit then invalid_arg ("fprintf: unknown format " ^ format)
-          else
-          begin match format.[j] with
-          | '@' ->
-              pp_print_char ppf '@';
-              doprn (succ j)
+          let i = succ i in
+          if i >= limit then invalid_format format i else
+          begin match format.[i] with
           | '[' ->
-              let j = do_pp_open ppf (i + 2) in
-              doprn j
+              do_pp_open_box ppf (succ i)
           | ']' ->
               pp_close_box ppf ();
-              doprn (succ j)
+              doprn (succ i)
+          | '{' ->
+              do_pp_open_tag ppf (succ i)
+          | '}' ->
+              pp_close_tag ppf ();
+              doprn (succ i)
           | ' ' ->
               pp_print_space ppf ();
-              doprn (succ j)
+              doprn (succ i)
           | ',' ->
               pp_print_cut ppf ();
-              doprn (succ j)
+              doprn (succ i)
           | '?' ->
               pp_print_flush ppf ();
-              doprn (succ j)
+              doprn (succ i)
           | '.' ->
               pp_print_newline ppf ();
-              doprn (succ j)
+              doprn (succ i)
           | '\n' ->
               pp_force_newline ppf ();
-              doprn (succ j)
+              doprn (succ i)
           | ';' ->
-              let j = do_pp_break ppf (i + 2) in
-              doprn j
+              do_pp_break ppf (succ i)
           | '<' ->
-              let size, j =
-                get_int "fprintf: bad print format " format (i + 2) in
-              if format.[pred j] != '>'
-               then invalid_arg ("fprintf: bad print format " ^ format)
-               else print_as := Some size;
-              doprn j
-          | c -> format_invalid_arg "fprintf: unknown format " c end
-      | '%' ->
-          let j = skip_args (succ i) in
-          begin match format.[j] with
-          | '%' ->
-              pp_print_char ppf '%';
-              doprn (succ j)
-          | 's' ->
-              Obj.magic(fun s ->
-                if j <= succ i then
-                  pp_print_as_string ppf s
-                else begin
-                  let p =
-                    try
-                      int_of_string (String.sub format (succ i) (j - i - 1))
-                    with _ ->
-                      invalid_arg ("fprintf: bad %s format, " ^ format) in
-                  if p > 0 && String.length s < p then begin
-                    pp_print_as_string ppf
-                                  (String.make (p - String.length s) ' ');
-                    pp_print_as_string ppf s
-                  end else if p < 0 && String.length s < -p then begin
-                    pp_print_as_string ppf s;
-                    pp_print_as_string ppf
-                                  (String.make (-p - String.length s) ' ')
-                  end else
-                    pp_print_as_string ppf s
-                end;
-                doprn (succ j))
-          | 'c' ->
-              Obj.magic(fun c ->
-                pp_print_as_char ppf c;
-                doprn (succ j))
-          | 'd' | 'i' | 'o' | 'x' | 'X' | 'u' ->
-              Obj.magic(fun n ->
-                pp_print_as_string ppf
-                 (format_int (String.sub format i (j - i + 1)) n);
-                doprn (succ j))
-          | 'f' | 'e' | 'E' | 'g' | 'G' ->
-              Obj.magic(fun f ->
-                pp_print_as_string ppf
-                 (format_float (String.sub format i (j - i + 1)) f);
-                doprn (succ j))
-          | 'b' ->
-              Obj.magic(fun b ->
-                pp_print_as_string ppf (string_of_bool b);
-                doprn(succ j))
-          | 'a' ->
-              if str then
-               Obj.magic(fun printer arg ->
-                pp_print_as_string ppf (printer () arg);
-                doprn(succ j))
-              else
-               Obj.magic(fun printer arg ->
-                printer ppf arg;
-                doprn(succ j))
-          | 't' ->
-              if str then
-               Obj.magic(fun printer ->
-                pp_print_as_string ppf (printer ());
-                doprn(succ j))
-              else
-               Obj.magic(fun printer ->
-                printer ppf;
-                doprn(succ j))
-          | c ->
-              format_invalid_arg "fprintf: unknown format " c
+              let got_size size i =
+                print_as := Some size;
+                doprn (skip_gt i) in
+              get_int (succ i) got_size
+          | '@' as c ->
+              pp_print_as_char c;
+              doprn (succ i)
+          | c -> invalid_format format i
           end
-       | c -> pp_print_as_char ppf c; doprn (succ i)
+      | c ->
+         pp_print_as_char c;
+         doprn (succ i)
 
-  and skip_args j =
-    match format.[j] with
-    | '0' .. '9' | ' ' | '.' | '-' -> skip_args (succ j)
-    | c -> j
+  and cont_s s i =
+    pp_print_as_string s; doprn i
+  and cont_a printer arg i =
+    if str then
+      pp_print_as_string ((Obj.magic printer) () arg)
+    else
+      printer ppf arg;
+    doprn i
+  and cont_t printer i =
+    if str then
+      pp_print_as_string ((Obj.magic printer) ())
+    else
+      printer ppf;
+    doprn i
 
-  and get_int s1 s2 i =
-   if i >= limit then invalid_arg (s1 ^ s2) else
+  and get_int i c =
+   if i >= limit then invalid_integer format i else
    match format.[i] with
-   | ' ' -> get_int s1 s2 (succ i)
-   | c ->
+   | ' ' -> get_int (succ i) c
+   | '%' ->
+      let cont_s s i = c (format_int_of_string format i s) i
+      and cont_a printer arg i = invalid_integer format i
+      and cont_t printer i = invalid_integer format i in
+      Printf.scan_format format i cont_s cont_a cont_t
+   | _ ->
       let rec get j =
-       if j >= limit then invalid_arg (s1 ^ s2) else
+       if j >= limit then invalid_integer format j else
        match format.[j] with
        | '0' .. '9' | '-' -> get (succ j)
-       | '>' | ' ' ->
-         if j = i then 0, succ j else
-          begin try int_of_string (String.sub format i (j - i)), succ j
-          with Failure _ -> invalid_arg (s1 ^ s2) end
-       | c -> format_invalid_arg (s1 ^ s2) c in
-       get i
+       | _ ->
+         if j = i then c 0 j else
+         c (format_int_of_string format j (String.sub format i (j - i))) j in
+      get i
 
-  and get_box_kind j =
-   if j >= limit then Pp_box, j else
-   match format.[j] with
+  and skip_gt i =
+   if i >= limit then invalid_format format i else
+   match format.[i] with
+   | ' ' -> skip_gt (succ i)
+   | '>' -> succ i
+   | _ -> invalid_format format i
+
+  and get_box_kind i =
+   if i >= limit then Pp_box, i else
+   match format.[i] with
    | 'h' ->
-      let j = succ j in
-      if j >= limit then Pp_hbox, j else
-      begin match format.[j] with
-      | 'o' -> 
-         let j = succ j in
-         if j >= limit
-          then invalid_arg ("fprintf: bad box format " ^ format) else
-         begin match format.[j] with
-         | 'v' -> Pp_hovbox, succ j
-         | c ->  format_invalid_arg "fprintf: bad name " c end
-      | 'v' -> Pp_hvbox, succ j
-      | c -> Pp_hbox, j
+      let i = succ i in
+      if i >= limit then Pp_hbox, i else
+      begin match format.[i] with
+      | 'o' ->
+         let i = succ i in
+         if i >= limit then format_invalid_arg "bad box format" format i else
+         begin match format.[i] with
+         | 'v' -> Pp_hovbox, succ i
+         | c ->
+            format_invalid_arg
+              ("bad box name ho" ^ String.make 1 c) format i end
+      | 'v' -> Pp_hvbox, succ i
+      | c -> Pp_hbox, i
       end
-   | 'b' -> Pp_box, succ j
-   | 'v' -> Pp_vbox, succ j
-   | _ -> Pp_box, j
+   | 'b' -> Pp_box, succ i
+   | 'v' -> Pp_vbox, succ i
+   | _ -> Pp_box, i
+
+  and get_tag_name i c =
+   let rec get accu i j =
+    if j >= limit
+    then c (implode_rev (String.sub format i (j - i)) accu) j else
+    match format.[j] with
+    | '>' -> c (implode_rev (String.sub format i (j - i)) accu) j
+    | '%' ->
+       let s0 = String.sub format i (j - i) in
+       let cont_s s i = get (s :: s0 :: accu) i i
+       and cont_a printer arg i =
+         let s =
+           if str then (Obj.magic printer) () arg else exstring printer arg in
+         get (s :: s0 :: accu) i i
+       and cont_t printer i =
+         let s =
+           if str then (Obj.magic printer) ()
+           else exstring (fun ppf () -> printer ppf) () in
+         get (s :: s0 :: accu) i i in
+       Printf.scan_format format j cont_s cont_a cont_t
+    | c -> get accu i (succ j) in
+   get [] i i
 
   and do_pp_break ppf i =
-   if i >= limit then begin pp_print_space ppf (); i end else
+   if i >= limit then begin pp_print_space ppf (); doprn i end else
    match format.[i] with
    | '<' ->
-     let nspaces, j =
-       get_int "fprintf: bad break format " format (succ i) in
-     let offset, j =
-       get_int "fprintf: bad break format " format j in
-     if format.[pred j] != '>'
-     then invalid_arg "fprintf: bad break format" format
-     else pp_print_break ppf nspaces offset;
-     j
-   | c ->  pp_print_space ppf (); i
+       let rec got_nspaces nspaces i =
+         get_int i (got_offset nspaces)
+       and got_offset nspaces offset i =
+         pp_print_break ppf nspaces offset;
+         doprn (skip_gt i) in
+       get_int (succ i) got_nspaces
+   | c -> pp_print_space ppf (); doprn i
 
-  and do_pp_open ppf i =
-   if i >= limit then begin pp_open_box_gen ppf 0 Pp_box; i end else
+  and do_pp_open_box ppf i =
+   if i >= limit then begin pp_open_box_gen ppf 0 Pp_box; doprn i end else
    match format.[i] with
    | '<' ->
-     let k, j = get_box_kind (succ i) in
-     let size, j = get_int "fprintf: bad box format " format j in
-     pp_open_box_gen ppf size k;
-     j
-   | c ->  pp_open_box_gen ppf 0 Pp_box; i
+     let kind, i = get_box_kind (succ i) in
+     let got_size size i =
+       pp_open_box_gen ppf size kind;
+       doprn (skip_gt i) in
+     get_int i got_size
+   | c -> pp_open_box_gen ppf 0 Pp_box; doprn i
 
-  in doprn 0;;
+  and do_pp_open_tag ppf i =
+   if i >= limit then begin pp_open_tag ppf ""; doprn i end else
+   match format.[i] with
+   | '<' ->
+     let got_name tag_name i =
+       pp_open_tag ppf tag_name;
+       doprn (skip_gt i) in
+     get_tag_name (succ i) got_name
+   | c -> pp_open_tag ppf ""; doprn i in
 
-let unit_out () = ();;
+  doprn 0;;
 
-let get_buffer_out b =
- let s = Buffer.contents b in
- Buffer.reset b;
- s;;
+(**************************************************************
 
-let string_out b ppf () =
- pp_flush_queue ppf false;
- get_buffer_out b;;
+  Defining [fprintf] and various flavors of [fprintf].
+
+ **************************************************************)
 
 let fprintf ppf = fprintf_out false unit_out ppf;;
-let printf f = fprintf_out false unit_out std_formatter f;;
-let eprintf f = fprintf_out false unit_out err_formatter f;;
-let sprintf f =
+let printf f = fprintf std_formatter f;;
+let eprintf f = fprintf err_formatter f;;
+
+let kprintf k =
  let b = Buffer.create 512 in
  let ppf = formatter_of_buffer b in
- fprintf_out true (string_out b ppf) ppf f;;
+ fprintf_out true (fun () -> k (string_out b ppf)) ppf;;
+let sprintf f = kprintf (fun x -> x) f;;
 
 let bprintf b =
  let ppf = formatter_of_buffer b in
  fprintf_out false (fun () -> pp_flush_queue ppf false) ppf;;
 
-let _ = at_exit print_flush;;
+at_exit print_flush;;
