@@ -6,11 +6,12 @@
 (*                                                                     *)
 (*  Copyright 1996 Institut National de Recherche en Informatique et   *)
 (*  en Automatique.  All rights reserved.  This file is distributed    *)
-(*  under the terms of the GNU Library General Public License.         *)
+(*  under the terms of the GNU Library General Public License, with    *)
+(*  the special exception on linking described in file ../LICENSE.     *)
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: int64.ml,v 1.6 2000/04/18 08:51:28 xleroy Exp $ *)
+(* $Id: int64.ml,v 1.9 2002/06/26 09:13:58 xleroy Exp $ *)
 
 (* Module [Int64]: 64-bit integers *)
 
@@ -51,3 +52,10 @@ external format : string -> int64 -> string = "int64_format"
 let to_string n = format "%d" n
 
 external of_string : string -> int64 = "int64_of_string"
+
+external bits_of_float : float -> int64 = "int64_bits_of_float"
+external float_of_bits : int64 -> float = "int64_float_of_bits"
+
+type t = int64
+
+let compare = (Pervasives.compare: t -> t -> int)
