@@ -6,11 +6,12 @@
 (*                                                                     *)
 (*  Copyright 1996 Institut National de Recherche en Informatique et   *)
 (*  en Automatique.  All rights reserved.  This file is distributed    *)
-(*  under the terms of the GNU Library General Public License.         *)
+(*  under the terms of the GNU Library General Public License, with    *)
+(*  the special exception on linking described in file ../LICENSE.     *)
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: pervasives.ml,v 1.38 1999/11/30 16:06:56 garrigue Exp $ *)
+(* $Id: pervasives.ml,v 1.52 2002/07/12 09:47:54 xleroy Exp $ *)
 
 (* type 'a option = None | Some of 'a *)
 
@@ -22,7 +23,6 @@ let failwith s = raise(Failure s)
 let invalid_arg s = raise(Invalid_argument s)
 
 exception Exit
-exception Assert_failure of (string * int * int)
 
 (* Comparisons *)
 
@@ -107,6 +107,32 @@ external float : int -> float = "%floatofint"
 external float_of_int : int -> float = "%floatofint"
 external truncate : float -> int = "%intoffloat"
 external int_of_float : float -> int = "%intoffloat"
+external float_of_bytes : string -> float = "float_of_bytes"
+let infinity =
+  float_of_bytes "\127\240\000\000\000\000\000\000"
+              (* 0x7F F0 00 00 00 00 00 00 *)
+let neg_infinity =
+  float_of_bytes "\255\240\000\000\000\000\000\000"
+              (* 0xFF F0 00 00 00 00 00 00 *)
+let nan =
+  float_of_bytes "\127\240\000\000\000\000\000\001"
+              (* 0x7F F0 00 00 00 00 00 01 *)
+let max_float =
+  float_of_bytes "\127\239\255\255\255\255\255\255"
+              (* 0x7f ef ff ff ff ff ff ff *)
+let min_float =
+  float_of_bytes "\000\016\000\000\000\000\000\000"
+              (* 0x00 10 00 00 00 00 00 00 *)
+let epsilon_float =
+  float_of_bytes "\060\176\000\000\000\000\000\000"
+              (* 0x3c b0 00 00 00 00 00 00 *)
+type fpclass =
+    FP_normal
+  | FP_subnormal
+  | FP_zero
+  | FP_infinite
+  | FP_nan
+external classify_float: float -> fpclass = "classify_float"
 
 (* String operations -- more in module String *)
 
@@ -127,7 +153,7 @@ let (^) s1 s2 =
 external int_of_char : char -> int = "%identity"
 external unsafe_char_of_int : int -> char = "%identity"
 let char_of_int n =
-  if n < 0 or n > 255 then invalid_arg "char_of_int" else unsafe_char_of_int n
+  if n < 0 || n > 255 then invalid_arg "char_of_int" else unsafe_char_of_int n
 
 (* Unit operations *)
 
@@ -155,8 +181,20 @@ let string_of_int n =
 
 external int_of_string : string -> int = "int_of_string"
 
+module String = struct
+  external get : string -> int -> char = "%string_safe_get"
+end
+
 let string_of_float f =
-  format_float "%.12g" f
+  let s = format_float "%.12g" f in
+  let l = string_length s in
+  let rec loop i =
+    if i >= l then s ^ "." else
+    if s.[i] = '.' || s.[i] = 'e' then s
+    else loop (i+1)
+  in
+  loop 0
+;;
 
 external float_of_string : string -> float = "float_of_string"
 
@@ -172,8 +210,8 @@ let rec (@) l1 l2 =
 type in_channel
 type out_channel
 
-external open_descriptor_out: int -> out_channel = "caml_open_descriptor"
-external open_descriptor_in: int -> in_channel = "caml_open_descriptor"
+external open_descriptor_out: int -> out_channel = "caml_open_descriptor_out"
+external open_descriptor_in: int -> in_channel = "caml_open_descriptor_in"
 
 let stdin = open_descriptor_in 0
 let stdout = open_descriptor_out 1
@@ -197,9 +235,17 @@ let open_out name =
 let open_out_bin name =
   open_out_gen [Open_wronly; Open_creat; Open_trunc; Open_binary] 0o666 name
 
-external fflush : out_channel -> unit = "caml_flush"
 external flush : out_channel -> unit = "caml_flush"
 
+external out_channels_list : unit -> out_channel list 
+                           = "caml_out_channels_list"
+
+let flush_all () = 
+  let rec iter = function
+      [] -> ()
+    | a::l -> flush a; iter l
+  in iter (out_channels_list ())
+
 external unsafe_output : out_channel -> string -> int -> int -> unit
                        = "caml_output"
 
@@ -209,7 +255,7 @@ let output_string oc s =
   unsafe_output oc s 0 (string_length s)
 
 let output oc s ofs len =
-  if ofs < 0 or ofs + len > string_length s
+  if ofs < 0 || len < 0 || ofs > string_length s - len
   then invalid_arg "output"
   else unsafe_output oc s ofs len
 
@@ -245,7 +291,7 @@ external unsafe_input : in_channel -> string -> int -> int -> int
                       = "caml_input"
 
 let input ic s ofs len =
-  if ofs < 0 or ofs + len > string_length s
+  if ofs < 0 || len < 0 || ofs > string_length s - len
   then invalid_arg "input"
   else unsafe_input ic s ofs len
 
@@ -258,29 +304,39 @@ let rec unsafe_really_input ic s ofs len =
   end
 
 let really_input ic s ofs len =
-  if ofs < 0 or ofs + len > string_length s
+  if ofs < 0 || len < 0 || ofs > string_length s - len
   then invalid_arg "really_input"
   else unsafe_really_input ic s ofs len
 
 external input_scan_line : in_channel -> int = "caml_input_scan_line"
 
-let rec input_line chan =
-  let n = input_scan_line chan in
-  if n = 0 then                         (* n = 0: we are at EOF *)
-    raise End_of_file
-  else if n > 0 then begin              (* n > 0: newline found in buffer *)
-    let res = string_create (n-1) in
-    ignore (unsafe_input chan res 0 (n-1));
-    ignore (input_char chan);           (* skip the newline *)
-    res
-  end else begin                        (* n < 0: newline not found *)
-    let beg = string_create (-n) in
-    ignore(unsafe_input chan beg 0 (-n));
-    try
-      beg ^ input_line chan
-    with End_of_file ->
-      beg
-  end
+let input_line chan =
+  let rec build_result buf pos = function
+    [] -> buf
+  | hd :: tl ->
+      let len = string_length hd in
+      string_blit hd 0 buf (pos - len) len;
+      build_result buf (pos - len) tl in
+  let rec scan accu len =
+    let n = input_scan_line chan in
+    if n = 0 then begin                   (* n = 0: we are at EOF *)
+      match accu with
+        [] -> raise End_of_file
+      | _  -> build_result (string_create len) len accu
+    end else if n > 0 then begin          (* n > 0: newline found in buffer *)
+      let res = string_create (n-1) in
+      ignore (unsafe_input chan res 0 (n-1));
+      ignore (input_char chan);           (* skip the newline *)
+      match accu with
+        [] -> res
+      |  _ -> let len = len + n - 1 in 
+              build_result (string_create len) len (res :: accu)
+    end else begin                        (* n < 0: newline not found *)
+      let beg = string_create (-n) in
+      ignore(unsafe_input chan beg 0 (-n));
+      scan (beg :: accu) (len - n)
+    end
+  in scan [] 0
 
 external input_byte : in_channel -> int = "caml_input_char"
 external input_binary_int : in_channel -> int = "caml_input_int"
@@ -317,6 +373,18 @@ let read_line () = flush stdout; input_line stdin
 let read_int () = int_of_string(read_line())
 let read_float () = float_of_string(read_line())
 
+(* Operations on large files *)
+
+module LargeFile =
+  struct
+    external seek_out : out_channel -> int64 -> unit = "caml_seek_out_64"
+    external pos_out : out_channel -> int64 = "caml_pos_out_64"
+    external out_channel_length : out_channel -> int64 = "caml_channel_size_64"
+    external seek_in : in_channel -> int64 -> unit = "caml_seek_in_64"
+    external pos_in : in_channel -> int64 = "caml_pos_in_64"
+    external in_channel_length : in_channel -> int64 = "caml_channel_size_64"
+  end
+
 (* References *)
 
 type 'a ref = { mutable contents: 'a }
@@ -330,7 +398,7 @@ external decr: int ref -> unit = "%decr"
 
 external sys_exit : int -> 'a = "sys_exit"
 
-let exit_function = ref (fun () -> flush stdout; flush stderr)
+let exit_function = ref flush_all
 
 let at_exit f =
   let g = !exit_function in
@@ -342,3 +410,6 @@ let exit retcode =
   do_at_exit ();
   sys_exit retcode
 
+external register_named_value: string -> 'a -> unit = "register_named_value"
+
+let _ = register_named_value "Pervasives.do_at_exit" do_at_exit
