@@ -6,18 +6,19 @@
 (*                                                                     *)
 (*  Copyright 1996 Institut National de Recherche en Informatique et   *)
 (*  en Automatique.  All rights reserved.  This file is distributed    *)
-(*  under the terms of the GNU Library General Public License.         *)
+(*  under the terms of the GNU Library General Public License, with    *)
+(*  the special exception on linking described in file ../LICENSE.     *)
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: string.ml,v 1.19 2000/01/12 15:53:18 doligez Exp $ *)
+(* $Id: string.ml,v 1.24 2002/07/12 09:47:54 xleroy Exp $ *)
 
 (* String operations *)
 
 external length : string -> int = "%string_length"
 external get : string -> int -> char = "%string_safe_get"
 external set : string -> int -> char -> unit = "%string_safe_set"
-external create: int -> string = "create_string"
+external create : int -> string = "create_string"
 external unsafe_get : string -> int -> char = "%string_unsafe_get"
 external unsafe_set : string -> int -> char -> unit = "%string_unsafe_set"
 external unsafe_blit : string -> int -> string -> int -> int -> unit
@@ -37,7 +38,7 @@ let copy s =
   r
 
 let sub s ofs len =
-  if ofs < 0 or len < 0 or ofs + len > length s
+  if ofs < 0 || len < 0 || ofs > length s - len
   then invalid_arg "String.sub"
   else begin
     let r = create len in
@@ -46,16 +47,19 @@ let sub s ofs len =
   end
 
 let fill s ofs len c =
-  if ofs < 0 or len < 0 or ofs + len > length s
+  if ofs < 0 || len < 0 || ofs > length s - len
   then invalid_arg "String.fill"
   else unsafe_fill s ofs len c
 
 let blit s1 ofs1 s2 ofs2 len =
-  if len < 0 or ofs1 < 0 or ofs1 + len > length s1
-             or ofs2 < 0 or ofs2 + len > length s2
+  if len < 0 || ofs1 < 0 || ofs1 > length s1 - len
+             || ofs2 < 0 || ofs2 > length s2 - len
   then invalid_arg "String.blit"
   else unsafe_blit s1 ofs1 s2 ofs2 len
 
+let iter f a =
+  for i = 0 to length a - 1 do f(unsafe_get a i) done
+
 let concat sep l =
   match l with
     [] -> ""
@@ -167,3 +171,7 @@ let rcontains_from s i c =
   try ignore(rindex_rec s i c); true with Not_found -> false;;
 
 let contains s c = contains_from s 0 c;;
+
+type t = string
+
+let compare = (Pervasives.compare: t -> t -> int)
