@@ -6,18 +6,19 @@
 (*                                                                     *)
 (*  Copyright 1996 Institut National de Recherche en Informatique et   *)
 (*  en Automatique.  All rights reserved.  This file is distributed    *)
-(*  under the terms of the GNU Library General Public License.         *)
+(*  under the terms of the GNU Library General Public License, with    *)
+(*  the special exception on linking described in file ../LICENSE.     *)
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: sys.ml,v 1.20 1999/11/23 10:50:06 xleroy Exp $ *)
+(* $Id: sys.ml,v 1.37 2002/08/05 14:16:12 xleroy Exp $ *)
 
 (* System interface *)
 
 external get_config: unit -> string * int = "sys_get_config"
-external get_argv: unit -> string array = "sys_get_argv"
+external get_argv: unit -> string * string array = "sys_get_argv"
 
-let argv = get_argv()
+let (executable_name, argv) = get_argv()
 let (os_type, word_size) = get_config()
 let max_array_length = (1 lsl (word_size - 10)) - 1;;
 let max_string_length = word_size / 8 * max_array_length - 1;;
@@ -72,3 +73,9 @@ let catch_break on =
     set_signal sigint (Signal_handle(fun _ -> raise Break))
   else
     set_signal sigint Signal_default
+
+
+(* OCaml version string, moved from utils/config.mlp.
+   Must be in the format described in sys.mli. *)
+
+let ocaml_version = "3.06"
